import six

import django
from django import forms
from django.conf import settings
from django.core.exceptions import ValidationError
from django.forms.models import ModelChoiceIterator
from django.forms.models import ChoiceField, ModelMultipleChoiceField
from django.utils.html import escape, conditional_escape

try:
    # Django 1.8+
    from django.forms.utils import flatatt
except ImportError:
    from django.forms.util import flatatt

try:
    from django.utils.encoding import force_unicode
except ImportError:
    from django.utils.encoding import force_text as force_unicode

from generic_plus.forms import BaseGenericFileInlineFormSet, GenericForeignFileWidget

from .utils import json


__all__ = ('CropDusterWidget', 'CropDusterThumbFormField', 'CropDusterInlineFormSet')


class CropDusterWidget(GenericForeignFileWidget):

    sizes = None

    template = "cropduster/custom_field.html"

    class Media:
        css = {'all': (u'%scropduster/css/cropduster.css?v=6' % settings.STATIC_URL,)}
        js = (
            u'%scropduster/js/jsrender.js' % settings.STATIC_URL,
            u'%scropduster/js/cropduster.js?v=8' % settings.STATIC_URL,
        )

    def get_context_data(self, name, value, attrs=None, bound_field=None):
        ctx = super(CropDusterWidget, self).get_context_data(name, value, attrs, bound_field)
        sizes = self.sizes
        if six.callable(sizes):
            instance = getattr(getattr(bound_field, 'form', None), 'instance', None)
            related_object = ctx['instance']
            sizes_callable = getattr(sizes, 'im_func', sizes)
            sizes = sizes_callable(instance, related=related_object)
        ctx.update({
            'sizes': json.dumps(sizes),
        })
        return ctx


class ThumbChoiceIterator(ModelChoiceIterator):

    def __iter__(self):
        if self.field.empty_label is not None:
            yield ("", self.field.empty_label)
        if getattr(self.field, 'cache_choices', None):
            if self.field.choice_cache is None:
                self.field.choice_cache = [
                    self.choice(obj) for obj in self.queryset
                ]
            for choice in self.field.choice_cache:
                yield choice
        else:
            for obj in self.queryset:
                yield self.choice(obj)

    def choice(self, obj):
        return (obj, self.field.label_from_instance(obj))


class CropDusterThumbWidget(forms.SelectMultiple):

    def __init__(self, *args, **kwargs):
        from cropduster.models import Thumb

        super(CropDusterThumbWidget, self).__init__(*args, **kwargs)
        self.model = Thumb

    def render_option(self, selected_choices, option_value, option_label):
        attrs = {}
        if isinstance(option_value, self.model):
            thumb = option_value
            option_value = thumb.pk
        else:
            try:
                thumb = self.model.objects.get(pk=option_value)
            except (TypeError, self.model.DoesNotExist):
                thumb = None

        if thumb:
            # If the thumb has no images associated with it then
            # it has not yet been saved, and so its file path has
            # '_tmp' appended before the extension.
            use_tmp_file = not(thumb.image_id)
            attrs = {
                'data-width': thumb.width,
                'data-height': thumb.height,
                'data-tmp-file': json.dumps(use_tmp_file),
            }
        option_value = force_unicode(option_value)
        if option_value in selected_choices:
            selected_html = u' selected="selected"'
            if not self.allow_multiple_selected:
                # Only allow for a single selection.
                selected_choices.remove(option_value)
        else:
            selected_html = ''
        return (
            u'<option value="%(value)s"%(selected)s%(attrs)s>%(label)s</option>') % {
                'value': escape(option_value),
                'selected': selected_html,
                'attrs': flatatt(attrs),
                'label': conditional_escape(force_unicode(option_label)),
        }


class CropDusterThumbFormField(ModelMultipleChoiceField):

    widget = CropDusterThumbWidget

    def clean(self, value):
        """
        Override default validation so that it doesn't throw a ValidationError
        if a given value is not in the original queryset.
        """
        try:
            value = super(CropDusterThumbFormField, self).clean(value)
        except ValidationError as e:
            if self.error_messages['required'] in e.messages:
                raise
            elif self.error_messages['list'] in e.messages:
                raise
        return value

    def _get_choices(self):
        if hasattr(self, '_choices'):
            return self._choices
        return ThumbChoiceIterator(self)

    choices = property(_get_choices, ChoiceField._set_choices)


def get_cropduster_field_on_model(model, field_identifier):
    from cropduster.fields import CropDusterField

    opts = model._meta
    if hasattr(opts, 'get_fields'):
        # Django 1.8+
        m2m_fields = [f for f in opts.get_fields() if f.many_to_many and not f.auto_created]
    else:
        m2m_fields = opts.many_to_many
    m2m_related_fields = set(m2m_fields + opts.virtual_fields)

    field_match = lambda f: (isinstance(f, CropDusterField)
        and f.field_identifier == field_identifier)

    try:
        return [f for f in m2m_related_fields if field_match(f)][0]
    except IndexError:
        return None


class CropDusterInlineFormSet(BaseGenericFileInlineFormSet):

    fields = ('image', 'thumbs', 'attribution', 'attribution_link', 'caption', 'field_identifier')

    def __init__(self, *args, **kwargs):
        super(CropDusterInlineFormSet, self).__init__(*args, **kwargs)
        if self.instance and not self.data:
            cropduster_field = get_cropduster_field_on_model(self.instance.__class__, self.field_identifier)
            if cropduster_field:
                # An order_by() is required to prevent the queryset result cache
                # from being removed
                self.queryset = self.queryset.order_by('pk')
                field_file = getattr(self.instance, cropduster_field.name)
                self.queryset._result_cache = list(filter(None, [field_file.related_object]))

    def _construct_form(self, i, **kwargs):
        """
        Limit the queryset of the thumbs for performance reasons (so that it doesn't
        pull in every available thumbnail into the selectbox)
        """
        from cropduster.models import Thumb

        form = super(CropDusterInlineFormSet, self)._construct_form(i, **kwargs)

        field_identifier_field = form.fields['field_identifier']
        field_identifier_field.widget = forms.HiddenInput()
        field_identifier_field.initial = self.field_identifier

        thumbs_field = form.fields['thumbs']

        if form.instance and form.instance.pk:
            # Set the queryset to the current list of thumbs on the image
            if django.VERSION < (1, 7):
                thumbs_field.queryset = form.instance.thumbs.get_query_set()
            else:
                thumbs_field.queryset = form.instance.thumbs.get_queryset()
        else:
            # Start with an empty queryset
            thumbs_field.queryset = Thumb.objects.none()

        if form.data:
            # Check if thumbs from POST data should be used instead.
            # These can differ from the values in the database if a
            # ValidationError elsewhere prevented saving.
            try:
                thumb_pks = [int(v) for v in form['thumbs'].value()]
            except (TypeError, ValueError):
                pass
            else:
                if thumb_pks and thumb_pks != [o.pk for o in thumbs_field.queryset]:
                    thumbs_field.queryset = Thumb.objects.filter(pk__in=thumb_pks)

        return form



from __future__ import absolute_import, division

import os
import PIL
import uuid
import shutil

from django import test
from django.contrib.contenttypes.models import ContentType

from .helpers import CropdusterTestCaseMediaMixin
from .models import Article, Author, TestForOptionalSizes, TestMultipleFieldsInheritanceChild
from ..models import Size, Image
from ..exceptions import CropDusterResizeException


class TestImage(CropdusterTestCaseMediaMixin, test.TestCase):

    def setUp(self):
        super(TestImage, self).setUp()
        author = Author.objects.create(name="Samuel Langhorne Clemens")
        article = Article.objects.create(title="Pudd'nhead Wilson",
            author=author, lead_image=os.path.join(self.TEST_IMG_DIR_RELATIVE, 'img.jpg'))
        article.lead_image.generate_thumbs()

        self.article_ct = ContentType.objects.get(app_label='cropduster', model='article')
        self.article = Article.objects.get(pk=article.pk)
        self.author = author

    def test_original_image(self):
        article = self.article
        self.assertEqual(article.lead_image.width, 674)
        self.assertEqual(article.lead_image.height, 800)
        self.assertIs(article.lead_image.sizes, Article.LEAD_IMAGE_SIZES)

    def test_generate_thumbs(self):
        article = self.article
        sizes = sorted(list(Size.flatten(Article.LEAD_IMAGE_SIZES)), key=lambda x: x.name)
        thumbs = sorted(list(article.lead_image.related_object.thumbs.all()), key=lambda x: x.name)
        self.assertEqual(len(thumbs), len(sizes))
        for size, thumb in zip(sizes, thumbs):
            self.assertEqual(size.name, thumb.name)
            if size.width:
                self.assertEqual(size.width, thumb.width)
            if size.height:
                self.assertEqual(size.height, thumb.height)
            else:
                ratio = article.lead_image.height / article.lead_image.width
                self.assertAlmostEqual(thumb.height, ratio * size.width, delta=1)
            self.assertEqual(thumb.image.image, article.lead_image.related_object.image)
            self.assertTrue(os.path.exists(thumb.path))
            self.assertEqual((thumb.width, thumb.height), PIL.Image.open(thumb.path).size)

        old_filepath = os.path.join(self.TEST_IMG_DIR, 'img.jpg')
        path, filename = os.path.split(old_filepath)
        new_filepath = os.path.join(path, 'new-%s' % filename)
        image = PIL.Image.open(old_filepath)
        image.thumbnail((50, 50))
        image.save(new_filepath)
        article = Article.objects.create(title="Img Too Small", author=self.author, lead_image=new_filepath)
        self.assertRaises(CropDusterResizeException, article.lead_image.generate_thumbs)

    def test_prefetch_related_with_images(self):
        for x in range(3):
            imgpath = os.path.join(self.TEST_IMG_DIR, '%s.jpg' % uuid.uuid4().hex)
            shutil.copyfile(os.path.join(self.TEST_IMG_DIR, 'img.jpg'), imgpath)
            article = Article.objects.create(title="", author=self.author, lead_image=imgpath)
            article.lead_image.generate_thumbs()

        with self.assertNumQueries(2):
            articles = Article.objects.filter(pk__in=[article.pk])
            for article in articles.prefetch_related('lead_image'):
                article.lead_image.related_object

        with self.assertNumQueries(3):
            articles = Article.objects.filter(pk__in=[article.pk])
            for article in articles.prefetch_related('lead_image__thumbs'):
                list(article.lead_image.related_object.thumbs.all())

    def test_redundant_prefetch_related_args_with_images(self):
        for x in range(3):
            imgpath = os.path.join(self.TEST_IMG_DIR, '%s.jpg' % uuid.uuid4().hex)
            shutil.copyfile(os.path.join(self.TEST_IMG_DIR, 'img.jpg'), imgpath)
            article = Article.objects.create(title="", author=self.author, lead_image=imgpath)
            article.lead_image.generate_thumbs()

        with self.assertNumQueries(3):
            articles = Article.objects.filter(pk__in=[article.pk])
            for article in articles.prefetch_related('lead_image', 'lead_image__thumbs'):
                list(article.lead_image.related_object.thumbs.all())

    def test_prefetch_related_with_alt_images(self):
        img_map = {}
        for x in range(3):
            img_name = uuid.uuid4().hex
            img_name_alt = uuid.uuid4().hex
            img_path = os.path.join(self.TEST_IMG_DIR, img_name, 'original.jpg')
            img_path_alt = os.path.join(self.TEST_IMG_DIR, img_name_alt, 'original.jpg')
            os.mkdir(os.path.dirname(img_path))
            os.mkdir(os.path.dirname(img_path_alt))
            shutil.copyfile(os.path.join(self.TEST_IMG_DIR, 'img.jpg'), img_path)
            shutil.copyfile(os.path.join(self.TEST_IMG_DIR, 'img.jpg'), img_path_alt)
            article = Article.objects.create(title="", author=self.author,
                lead_image=img_path, alt_image=img_path_alt)
            article.lead_image.generate_thumbs()
            article.alt_image.generate_thumbs()
            img_map[article.pk] = (img_name, img_name_alt)

        with self.assertNumQueries(5):
            articles = Article.objects.filter(id__in=img_map.keys()).prefetch_related(
               'lead_image__thumbs', 'alt_image__thumbs')
            for article in articles:
                lead_name, alt_name = img_map[article.pk]
                self.assertTrue(lead_name in article.lead_image.related_object.path)
                self.assertTrue(alt_name in article.alt_image.related_object.path)

                lead_sizes = [s.name for s in Size.flatten(Article.LEAD_IMAGE_SIZES)]
                lead_thumbs = list(article.lead_image.related_object.thumbs.all())
                self.assertEqual(len(lead_thumbs), len(lead_sizes))
                for thumb in lead_thumbs:
                    self.assertEqual(thumb.image_id, article.lead_image.related_object.pk)
                    self.assertTrue(thumb.name in lead_sizes)

                alt_sizes = [s.name for s in Size.flatten(Article.ALT_IMAGE_SIZES)]
                alt_thumbs = list(article.alt_image.related_object.thumbs.all())
                self.assertEqual(len(alt_thumbs), len(alt_sizes))
                for thumb in alt_thumbs:
                    self.assertEqual(thumb.image_id, article.alt_image.related_object.pk)
                    self.assertTrue(thumb.name in alt_sizes)

    def test_prefetch_related_through_table(self):
        author = Author.objects.create(name="Author")
        for x in range(10):
            imgpath = os.path.join(self.TEST_IMG_DIR, '%s.jpg' % uuid.uuid4().hex)
            shutil.copyfile(os.path.join(self.TEST_IMG_DIR, 'img.jpg'), imgpath)
            alt_imgpath = os.path.join(self.TEST_IMG_DIR, '%s.png' % uuid.uuid4().hex)
            shutil.copyfile(os.path.join(self.TEST_IMG_DIR, 'img.png'), alt_imgpath)
            article = Article.objects.create(title="prefetch", author=author, lead_image=imgpath, alt_image=alt_imgpath)
            article.lead_image.generate_thumbs()
            article.alt_image.generate_thumbs()

        lead_sizes = [s.name for s in Size.flatten(Article.LEAD_IMAGE_SIZES)]
        alt_sizes = [s.name for s in Size.flatten(Article.ALT_IMAGE_SIZES)]

        with self.assertNumQueries(6):
            authors = Author.objects.filter(pk=author.pk).prefetch_related(
                'article_set__lead_image__thumbs',
                'article_set__alt_image__thumbs')
            for author in authors:
                for article in author.article_set.all():
                    self.assertTrue(article.lead_image.path.endswith('jpg'))
                    self.assertTrue(article.alt_image.path.endswith('png'))
                    for thumb in article.lead_image.related_object.thumbs.all():
                        self.assertTrue(thumb.name in lead_sizes)
                    for thumb in article.alt_image.related_object.thumbs.all():
                        self.assertTrue(thumb.name in alt_sizes)


class TestModelSaving(CropdusterTestCaseMediaMixin, test.TestCase):

    def test_save_image_updates_model(self):
        img_path = os.path.join(self.TEST_IMG_DIR, 'img.png')
        article = Article.objects.create(title="test", author=Author.objects.create(name='test'))
        Image.objects.create(content_type=ContentType.objects.get(app_label='cropduster', model='article'),
                             object_id=article.pk,
                             image=img_path)
        self.assertFalse(article.lead_image)

        # Refresh the article from the database
        article = Article.objects.get(pk=article.pk)
        self.assertTrue(article.lead_image)
        self.assertEqual(article.lead_image.path, img_path)

    def test_resave_single_image_model(self):
        author = Author(name='test')
        author.headshot = os.path.join(self.TEST_IMG_DIR_RELATIVE, 'img.jpg')
        author.save()

        author = Author.objects.get(pk=author.pk)
        author.save()

        author = Author.objects.get(pk=author.pk)
        self.assertTrue(author.headshot.path.endswith('img.jpg'))

    def test_resave_single_image_model_with_thumbs(self):
        author = Author(name='test')
        author.headshot = os.path.join(self.TEST_IMG_DIR_RELATIVE, 'img.jpg')
        author.save()

        author.headshot.generate_thumbs()

        author = Author.objects.get(pk=author.pk)
        author.save()

        author = Author.objects.get(pk=author.pk)
        self.assertTrue(author.headshot.path.endswith('img.jpg'))

    def test_resave_multi_image_model_with_one_image(self):
        article = Article(title="title", author=Author.objects.create(name='test'))
        article.lead_image = os.path.join(self.TEST_IMG_DIR_RELATIVE, 'img.jpg')
        article.save()

        article = Article.objects.get(pk=article.pk)
        article.save()

        article = Article.objects.get(pk=article.pk)
        self.assertTrue(article.lead_image.path.endswith('img.jpg'))

    def test_resave_multi_image_model_with_one_image_and_thumbs(self):
        article = Article(title="title", author=Author.objects.create(name='test'))
        article.lead_image = os.path.join(self.TEST_IMG_DIR_RELATIVE, 'img.jpg')
        article.save()

        article.lead_image.generate_thumbs()

        article = Article.objects.get(pk=article.pk)
        article.save()

        article = Article.objects.get(pk=article.pk)
        self.assertTrue(article.lead_image.path.endswith('img.jpg'))

    def test_resave_multi_image_model_with_two_images(self):
        article = Article(title="title", author=Author.objects.create(name='test'))
        article.lead_image = os.path.join(self.TEST_IMG_DIR_RELATIVE, 'img.jpg')
        article.alt_image = os.path.join(self.TEST_IMG_DIR_RELATIVE, 'img.png')
        article.save()

        article = Article.objects.get(pk=article.pk)
        article.save()

        article = Article.objects.get(pk=article.pk)
        self.assertTrue(article.lead_image.path.endswith('img.jpg'))
        self.assertTrue(article.alt_image.path.endswith('img.png'))

    def test_resave_multi_image_model_with_two_images_and_thumbs(self):
        article = Article(title="title", author=Author.objects.create(name='test'))
        article.lead_image = os.path.join(self.TEST_IMG_DIR_RELATIVE, 'img.jpg')
        article.alt_image = os.path.join(self.TEST_IMG_DIR_RELATIVE, 'img.png')
        article.save()

        article.lead_image.generate_thumbs()
        article.alt_image.generate_thumbs()

        article = Article.objects.get(pk=article.pk)
        article.save()

        article = Article.objects.get(pk=article.pk)
        self.assertTrue(article.lead_image.path.endswith('img.jpg'))
        self.assertTrue(article.alt_image.path.endswith('img.png'))

    def test_change_image_on_model_with_two_images(self):
        article = Article(title="title", author=Author.objects.create(name='test'))
        article.lead_image = os.path.join(self.TEST_IMG_DIR_RELATIVE, 'img.jpg')
        article.save()

        article = Article.objects.get(pk=article.pk)
        article.alt_image = os.path.join(self.TEST_IMG_DIR_RELATIVE, 'img.png')
        article.save()

        article = Article.objects.get(pk=article.pk)
        self.assertTrue(article.lead_image.path.endswith('img.jpg'))
        self.assertTrue(article.alt_image.path.endswith('img.png'))

    def test_optional_sizes(self):
        test_a = TestForOptionalSizes.objects.create(slug='a')
        test_a.image = os.path.join(self.TEST_IMG_DIR_RELATIVE, 'img.jpg')
        test_a.save()
        test_a.image.generate_thumbs()

        image = Image.objects.get(content_type=ContentType.objects.get_for_model(test_a),
            object_id=test_a.pk)
        num_thumbs = len(image.thumbs.all())
        self.assertEqual(num_thumbs, 1, "Expected one thumb; instead got %d" % num_thumbs)

        test_b = TestForOptionalSizes.objects.create(slug='b')
        test_b.image = os.path.join(self.TEST_IMG_DIR_RELATIVE, 'img2.jpg')
        test_b.save()
        test_b.image.generate_thumbs()

        image = Image.objects.get(content_type=ContentType.objects.get_for_model(test_b),
            object_id=test_b.pk)
        num_thumbs = len(image.thumbs.all())
        self.assertEqual(num_thumbs, 2, "Expected one thumb; instead got %d" % num_thumbs)

    def test_multiple_fields_with_inheritance(self):
        child_fields = [f.name for f in TestMultipleFieldsInheritanceChild._meta.local_fields]
        self.assertNotIn('image', child_fields,
            "Field 'image' from parent model should not be in the child model's local_fields")



from __future__ import division

import six

from six.moves import xrange

import hashlib
import random
import os
from datetime import datetime

import django
from django.core.exceptions import ImproperlyConfigured, ObjectDoesNotExist
from django.core.files.storage import FileSystemStorage
from django.contrib.contenttypes.models import ContentType
from django.db import models

try:
    from django.contrib.contenttypes.fields import GenericForeignKey
except ImportError:
    from django.contrib.contenttypes.generic import GenericForeignKey

import PIL.Image

from generic_plus.utils import get_relative_media_url

from .exceptions import CropDusterResizeException
from .fields import (
    CropDusterField, ReverseForeignRelation, CropDusterImageField,
    CropDusterSimpleImageField)
from .files import VirtualFieldFile
from .resizing import Size, Box, Crop
from .utils import process_image
from . import settings as cropduster_settings


__all__ = ('Image', 'Thumb', 'StandaloneImage', 'CropDusterField', 'Size', 'Box', 'Crop')


def safe_str_path(file_path):
    """
    Convert unicode paths to bytestrings so that os.path does not throw
    string conversion errors
    """
    if six.PY2 and isinstance(file_path, unicode):
        return file_path.encode('utf-8')
    return file_path


class Thumb(models.Model):

    name = models.CharField(max_length=255, db_index=True)
    width = models.PositiveIntegerField(default=0, blank=True, null=True)
    height = models.PositiveIntegerField(default=0, blank=True, null=True)

    # For a given thumbnail, it either has crop data or it references
    # another thumbnail with crop data
    reference_thumb = models.ForeignKey('Thumb', blank=True, null=True,
            related_name='auto_set')

    crop_x = models.PositiveIntegerField(blank=True, null=True)
    crop_y = models.PositiveIntegerField(blank=True, null=True)
    crop_w = models.PositiveIntegerField(blank=True, null=True)
    crop_h = models.PositiveIntegerField(blank=True, null=True)

    date_modified = models.DateTimeField(auto_now=True)

    image = models.ForeignKey('Image', related_name='+', null=True, blank=True)

    class Meta:
        app_label = cropduster_settings.CROPDUSTER_APP_LABEL
        db_table = '%s_thumb' % cropduster_settings.CROPDUSTER_DB_PREFIX

    def __unicode__(self):
        return self.name

    @property
    def image_file(self):
        return Image.get_file_for_size(image=self.image, size_name=self.name)

    @property
    def url(self):
        return self.image_file.url if self.image_file else ''

    @property
    def path(self):
        return self.image_file.path if self.image_file else ''

    def save(self, *args, **kwargs):
        if self.pk:
            try:
                orig_thumb = Thumb.objects.select_for_update().get(pk=self.pk)
            except Thumb.DoesNotExist:
                pass
            else:
                if self.image_id and not orig_thumb.image_id:
                    try:
                        os.rename(
                            self.image.get_image_path(self.name, tmp=True),
                            self.image.get_image_path(self.name))
                    except (IOError, OSError):
                        pass
        return super(Thumb, self).save(*args, **kwargs)

    def to_dict(self):
        """Returns a dict of the thumb's values which are JSON serializable."""
        dct = {}
        for k, v in six.iteritems(vars(self)):
            if isinstance(v, (six.string_types, float, bool, type(None))):
                dct[k] = v
            if isinstance(v, six.integer_types):
                dct[k] = v
        return dct

    def get_crop_box(self):
        if self.reference_thumb:
            ref_thumb = self.reference_thumb
        else:
            ref_thumb = self
        x1, y1 = ref_thumb.crop_x, ref_thumb.crop_y
        x2, y2 = x1 + ref_thumb.crop_w, y1 + ref_thumb.crop_h
        if any([getattr(ref_thumb, 'crop_%s' % a) is None for a in ['x', 'y', 'w', 'h']]):
            return None
        return Box(x1, y1, x2, y2)

    def crop(self, output_filename, original_image=None, w=None, h=None, min_w=None, min_h=None, max_w=None, max_h=None):
        if original_image is None:
            if not self.pk:
                raise Exception(
                    u"The `original_image` argument is required for"
                    u" thumbnails which have not yet been saved")

            if not self.image_id:
                raise Exception(
                    u"The `original_image` argument is required for"
                    u" thumbnails which are not associated with an image")

            original_image = self.image

        crop_box = self.get_crop_box()
        if crop_box is None:
            raise Exception("Cannot crop thumbnail without crop data")
        crop = Crop(crop_box, original_image)

        self.width = w or None
        self.height = h or None

        if self.reference_thumb:
            best_fit_kwargs = {
                'min_w': min_w or self.width,
                'min_h': min_h or self.height,
            }
            if self.width and self.height:
                best_fit_kwargs.update({'w': self.width, 'h': self.height})
            fit = crop.best_fit(**best_fit_kwargs)
            if not self.width and not self.height:
                self.width, self.height = fit.box.size
            elif not self.width:
                width = fit.box.w * (self.height / fit.box.h)
                self.width = min(int(round(width)), crop.bounds.w)
            elif not self.height:
                height = fit.box.h * (self.width / fit.box.w)
                self.height = min(int(round(height)), crop.bounds.h)
            new_image = fit.create_image(output_filename, width=self.width, height=self.height, max_w=max_w, max_h=max_h)
        else:
            if w and h:
                self.width = w
                self.height = h
            elif w:
                height = crop_box.h * (w / crop_box.w)
                self.height = min(int(round(height)), crop.bounds.h)
            elif h:
                width = crop_box.w * (h / crop_box.h)
                self.width = min(int(round(width)), crop.bounds.w)
            else:
                self.width, self.height = crop.box.size

            new_image = crop.create_image(output_filename, width=self.width, height=self.height, max_w=max_w, max_h=max_h)

        self.width, self.height = new_image.size
        return new_image


class StrFileSystemStorage(FileSystemStorage):
    """Converts paths to byte-strings.

    Linux uses str/bytes for file paths, but Django tries to use unicode.
    """
    def path(self, name):
        path = super(StrFileSystemStorage, self).path(name)
        if six.PY2 and isinstance(path, unicode):
            path = path.encode('utf-8')
        return path


image_storage = StrFileSystemStorage()


def generate_filename(instance, filename):
    return filename


class Image(models.Model):

    content_type = models.ForeignKey(ContentType)
    object_id = models.PositiveIntegerField(null=True, blank=True)
    content_object = GenericForeignKey('content_type', 'object_id')
    field_identifier = models.SlugField(null=False, blank=True, default="")

    prev_object_id = models.PositiveIntegerField(null=True, blank=True)
    prev_content_object = GenericForeignKey('content_type', 'prev_object_id')

    width = models.PositiveIntegerField(blank=True, null=True)
    height = models.PositiveIntegerField(blank=True, null=True)

    image = CropDusterSimpleImageField(db_index=True,
        upload_to=generate_filename, db_column='path',
        storage=image_storage, width_field='width', height_field='height')

    thumbs = ReverseForeignRelation(Thumb, field_name='image')

    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)

    attribution = models.CharField(max_length=255, blank=True, null=True)
    attribution_link = models.URLField(max_length=255, blank=True, null=True)
    caption = models.TextField(blank=True, null=True)

    class Meta:
        app_label = cropduster_settings.CROPDUSTER_APP_LABEL
        db_table = '%s_image' % cropduster_settings.CROPDUSTER_DB_PREFIX
        unique_together = ("content_type", "object_id", "field_identifier")

    def __unicode__(self):
        return self.get_image_url()

    @property
    def path(self):
        return self.image.name if self.image else None

    @property
    def url(self):
        return self.image.url if self.image else None

    @property
    def extension(self):
        ''' returns the file extension with a dot (.) prepended to it '''
        if not self.image:
            return u''
        return os.path.splitext(safe_str_path(self.image.path))[1]

    @staticmethod
    def get_file_for_size(image, size_name='original', tmp=False):
        if isinstance(image, six.string_types):
            image = VirtualFieldFile(image)
        if not image:
            return None
        path, basename = os.path.split(safe_str_path(image.path))
        filename, extension = os.path.splitext(basename)
        if size_name == 'preview':
            size_name = '_preview'
        if tmp:
            size_name = '%s_tmp' % size_name
        return VirtualFieldFile(
            '/'.join([
                get_relative_media_url(path),
                safe_str_path(size_name) + extension]))

    @classmethod
    def save_preview_file(cls, image_file, preview_w=None, preview_h=None):
        pil_img = PIL.Image.open(safe_str_path(image_file.path))
        orig_w, orig_h = pil_img.size

        preview_w = preview_w or cropduster_settings.CROPDUSTER_PREVIEW_WIDTH
        preview_h = preview_h or cropduster_settings.CROPDUSTER_PREVIEW_HEIGHT

        resize_ratio = min(preview_w / orig_w, preview_h / orig_h)

        def fit_preview(im):
            if resize_ratio < 1:
                w = int(round(orig_w * resize_ratio))
                h = int(round(orig_h * resize_ratio))
                preview_img = im.resize((w, h), PIL.Image.ANTIALIAS)
            else:
                w, h = orig_w, orig_h
                preview_img = im
            return preview_img

        preview_file = cls.get_file_for_size(image_file, '_preview')
        process_image(pil_img, safe_str_path(preview_file.path), fit_preview)
        return preview_file

    def save_preview(self, preview_w=None, preview_h=None):
        return Image.save_preview_file(self.image, preview_w=preview_w, preview_h=preview_h)

    def has_thumb(self, size_name):
        try:
            self.thumbs.get(name=size_name)
        except Thumb.DoesNotExist:
            return False
        else:
            return True

    def get_image_filesize(self, size_name='original'):
        size_name = size_name or 'original'
        if size_name != 'original' and not self.has_thumb(size_name):
            return 0
        return os.path.getsize(self.get_image_path(size_name))

    def get_image_filename(self, size_name='original'):
        size_name = size_name or 'original'
        if size_name != 'original' and not self.has_thumb(size_name):
            return ''
        return os.path.basename(self.get_image_path(size_name))

    def get_image_path(self, size_name='original', tmp=False):
        size_name = size_name or 'original'
        converted = Image.get_file_for_size(self.image, size_name, tmp=tmp)
        if not converted:
            return u''
        else:
            return converted.path

    def save(self, **kwargs):
        self.date_modified = datetime.now()
        if self.field_identifier is None:
            self.field_identifier = ""
        if not self.pk and self.content_type and self.object_id:
            try:
                original = Image.objects.get(content_type=self.content_type,
                                             object_id=self.object_id,
                                             field_identifier=self.field_identifier,
                                             prev_object_id__isnull=True)
            except Image.DoesNotExist:
                pass
            else:
                original.prev_object_id = original.object_id
                original.object_id = None
                original.save()

        super(Image, self).save(**kwargs)

        # If the Image has changed, we need to make sure the related field on the
        # model class has also been updated
        model_class = self.content_type.model_class()

        if django.VERSION >= (1, 8):
            fields_with_models = [
                (f, f.model if f.model != model_class else None)
                for f in model_class._meta.get_fields()
                if not f.is_relation
                    or f.one_to_one
                    or (f.many_to_one and f.related_model)
            ]
        else:
            fields_with_models = model_class._meta.get_fields_with_model()

        for field, field_model_class in fields_with_models:
            field_model_class = field_model_class or model_class
            if (isinstance(field, CropDusterImageField) and
                    field.generic_field.field_identifier == self.field_identifier):
                field_model_class.objects.filter(pk=self.object_id).update(**{field.attname: self.path or ''})

    def get_image_url(self, size_name='original', tmp=False):
        converted = Image.get_file_for_size(self.image, size_name, tmp=tmp)
        return getattr(converted, 'url', None) or u''

    def get_image_size(self, size_name=None):
        """
        Returns tuple of a thumbnail's size (width, height).
        When first parameter unspecified returns a tuple of the size of
        the original image.
        """
        if size_name is not None:
            try:
                thumb = self.thumbs.get(name=size_name)
            except Thumb.DoesNotExist:
                return (0, 0)
            else:
                return (thumb.width, thumb.height)

        # Get the original size
        if not self.image or not os.path.exists(safe_str_path(self.image.path)):
            return (0, 0)
        elif self.width and self.height:
            return (self.width, self.height)
        else:
            try:
                img = PIL.Image.open(safe_str_path(self.image.path))
            except (IOError, ValueError, TypeError):
                return (0, 0)
            else:
                return img.size

    def delete(self, *args, **kwargs):
        obj = self.content_object
        image_name = self.image.name if (self.image) else None

        super(Image, self).delete(*args, **kwargs)

        if not obj or not image_name:
            return

        def field_matches(f):
            if not isinstance(f, CropDusterImageField):
                return False
            obj_image_name = getattr(getattr(obj, f.name, None), 'name', None)
            return (obj_image_name == image_name)

        try:
            cropduster_field = [f for f in obj._meta.fields if field_matches(f)][0]
        except IndexError:
            pass
        else:
            # Clear the file field on the generic-related instance
            setattr(obj, cropduster_field.name, None)
            obj.save()

    def save_size(self, size, thumb=None, image=None, tmp=False, standalone=False, permissive=False):
        thumbs = {}
        if not image and not self.image:
            raise Exception("Cannot save sizes without an image")

        image = image or PIL.Image.open(safe_str_path(self.image.path))

        if standalone:
            if not StandaloneImage:
                raise ImproperlyConfigured(u"standalone mode used, but not installed.")
            return self._save_thumb(size, image, thumb, standalone=True)

        for sz in Size.flatten([size]):
            try:
                if sz.is_auto:
                    new_thumb = self._save_thumb(sz, image, ref_thumb=thumb, tmp=tmp)
                else:
                    thumb = new_thumb = self._save_thumb(sz, image, thumb, tmp=tmp)
            except CropDusterResizeException:
                if permissive or not sz.required:
                    continue
                else:
                    raise

            if new_thumb:
                thumbs[sz.name] = new_thumb
        return thumbs

    def _save_thumb(self, size, image=None, thumb=None, ref_thumb=None, tmp=False, standalone=False):
        img_save_params = {}
        if image.format == 'JPEG':
            img_save_params['quality'] = cropduster_settings.get_jpeg_quality(self.width, self.height)
        if image.format in ('JPEG', 'PNG') and cropduster_settings.JPEG_SAVE_ICC_SUPPORTED:
            img_save_params['icc_profile'] = image.info.get('icc_profile')

        if not thumb:
            if standalone:
                thumb = Thumb(
                    width=self.width, height=self.height,
                    crop_x=0, crop_y=0, crop_w=self.width, crop_h=self.height)
            elif self.pk:
                try:
                    thumb = self.thumbs.get(name=size.name)
                except Thumb.DoesNotExist:
                    pass

        if standalone:
            thumb.name = ''.join([random.choice('abcdefghijklmnopqrstuvwxyz0123456789') for i in xrange(0, 8)])
        elif not thumb:
            thumb = Thumb(name=size.name)
        elif not thumb.name:
            thumb.name = size.name

        crop_box = None

        if size.is_auto:
            ref_thumb = ref_thumb or thumb.reference_thumb
            if ref_thumb:
                crop_box = ref_thumb.get_crop_box()
        elif thumb:
            crop_box = thumb.get_crop_box()

        if not crop_box:
            return None

        thumb.reference_thumb = ref_thumb

        crop_kwargs = dict([(k, getattr(size, k))
                            for k in ['w', 'h', 'min_w', 'min_h', 'max_w', 'max_h']])
        if standalone and not(size.w or size.h) and (thumb.crop_w and thumb.crop_h):
            crop_kwargs['w'] = thumb.crop_w
            crop_kwargs['h'] = thumb.crop_h

        if standalone:
            thumb_path = self.get_image_path(thumb.name)
        else:
            thumb_path = self.get_image_path(size.name, tmp=tmp)

        thumb_image = thumb.crop(thumb_path, image, **crop_kwargs)

        if StandaloneImage:
            thumb_image.crop.add_xmp_to_crop(thumb_path, size, original_image=image)

        if standalone:
            md5 = hashlib.md5()
            with open(thumb_path, mode='rb') as f:
                md5.update(f.read())
            thumb.name = md5.hexdigest()[0:9]
            os.rename(thumb_path, self.get_image_path(thumb.name))
        else:
            thumb.save()
        return thumb


try:
    from cropduster.standalone.models import StandaloneImage
except:
    raise
    class FalseMeta(type):
        def __nonzero__(cls): return False
        __bool__ = __nonzero__

    @six.add_metaclass(FalseMeta)
    class StandaloneImage(object):
        DoesNotExist = type('DoesNotExist', (ObjectDoesNotExist,), {})


try:
    from south.modelsinspector import add_introspection_rules, add_ignored_fields
except ImportError:
    pass
else:
    add_ignored_fields(["^cropduster\.fields\.ReverseForeignRelation"])

    def converter(value):
        """Custom south converter so that Size objects serialize properly"""
        if isinstance(value, Size):
            return value.__serialize__()
        try:
            is_sizes_list = all([isinstance(sz, Size) for sz in value])
        except TypeError:
            pass
        else:
            if is_sizes_list:
                return [sz.__serialize__() for sz in value]
        return repr(value)

    add_introspection_rules(rules=[
        (
            (CropDusterField,),
            [],
            {
                "to": ["rel.to", {}],
                "symmetrical": ["rel.symmetrical", {"default": True}],
                "object_id_field": ["object_id_field_name", {"default": "object_id"}],
                "content_type_field": ["content_type_field_name", {"default": "content_type"}],
                "blank": ["blank", {"default": True}],
                "sizes": ["sizes", {"converter": converter}],
            },
        ),
    ], patterns=["^cropduster\.fields\.CropDusterField"])

    add_introspection_rules([], ["^cropduster\.fields\.CropDusterSimpleImageField"])



try:
    from django.conf.urls import url
except ImportError:
    from django.conf.urls.defaults import url

import cropduster.views
import cropduster.standalone.views


urlpatterns = [
    url(r'^$', cropduster.views.index, name='cropduster-index'),
    url(r'^crop/', cropduster.views.crop, name='cropduster-crop'),
    url(r'^upload/', cropduster.views.upload, name='cropduster-upload'),
    url(r'^standalone/', cropduster.standalone.views.index, name='cropduster-standalone'),
]



from __future__ import division

import os
import re
import hashlib

from six.moves.urllib import parse as urlparse
from six.moves.urllib.request import urlopen

import PIL.Image

from django.core.files.uploadedfile import SimpleUploadedFile
from django.conf import settings
from django.db.models.fields.files import FieldFile, FileField
from django.utils.functional import cached_property
from django.utils.http import urlunquote_plus

from generic_plus.utils import get_relative_media_url, get_media_path


class VirtualFieldFile(FieldFile):

    def __init__(self, name, storage=None, upload_to=None):
        super(FieldFile, self).__init__(None, name)
        self.instance = None
        self.field = FileField(name='file', upload_to=upload_to, storage=storage)
        self.storage = self.field.storage
        self._committed = True

    def get_directory_name(self):
        return self.field.get_directory_name()

    def get_filename(self, filename):
        return self.field.get_filename(filename)

    def generate_filename(self, filename):
        return self.field.generate_filename(None, filename)

    def save(self, *args, **kwargs):
        raise NotImplementedError

    def delete(self, *args, **kwargs):
        raise NotImplementedError

    @cached_property
    def dimensions(self):
        try:
            pil_image = PIL.Image.open(self.path)
        except:
            return (0, 0)
        else:
            return pil_image.size

    @cached_property
    def width(self):
        w, h = self.dimensions
        return w

    @cached_property
    def height(self):
        w, h = self.dimensions
        return h


class ImageFile(VirtualFieldFile):

    _path = None

    preview_image = None
    metadata = None

    def __init__(self, path, upload_to=None, preview_w=None, preview_h=None):
        self.upload_to = upload_to
        self.preview_width = preview_w
        self.preview_height = preview_h
        self.metadata = {}

        if not path:
            self.name = None
            return

        if '%' in path:
            path = urlunquote_plus(path)

        if path.startswith(settings.MEDIA_URL):
            # Strips leading MEDIA_URL, if starts with
            self._path = get_relative_media_url(path, clean_slashes=False)
        elif re.search(r'^(?:http(?:s)?:)?//', path):
            # url on other server? download it.
            self._path = self.download_image_url(path)
        else:
            abs_path = get_media_path(path)
            if os.path.exists(abs_path):
                self._path = get_relative_media_url(abs_path)

        if not self._path or not os.path.exists(os.path.join(settings.MEDIA_ROOT, self._path)):
            self.name = None
            return

        super(ImageFile, self).__init__(self._path)

        if self:
            self.preview_image = self.get_for_size('preview')

    def download_image_url(self, url):
        from cropduster.models import StandaloneImage
        from cropduster.views.forms import clean_upload_data

        image_contents = urlopen(url).read()
        md5_hash = hashlib.md5()
        md5_hash.update(image_contents)
        try:
            standalone_image = StandaloneImage.objects.get(md5=md5_hash.hexdigest())
        except StandaloneImage.DoesNotExist:
            pass
        else:
            return get_relative_media_url(standalone_image.image.name)

        parse_result = urlparse.urlparse(url)

        fake_upload = SimpleUploadedFile(os.path.basename(parse_result.path), image_contents)
        file_data = clean_upload_data({
            'image': fake_upload,
            'upload_to': self.upload_to,
        })
        return get_relative_media_url(file_data['image'].name)

    def __nonzero__(self):
        """When evaluated as boolean, base on whether self._path is not None"""
        if not self._path:
            return False
        return super(ImageFile, self).__nonzero__()

    def get_for_size(self, size_slug='original'):
        from cropduster.models import Image

        image = Image.get_file_for_size(self, size_slug)
        if size_slug == 'preview':
            if not os.path.exists(image.path):
                Image.save_preview_file(self, preview_w=self.preview_width, preview_h=self.preview_height)
        return image



import os

from django import test

from .helpers import CropdusterTestCaseMediaMixin
from cropduster.resizing import Crop, Box, Size


class TestResizing(CropdusterTestCaseMediaMixin, test.TestCase):

    def test_off_by_one_bug(self):
        img_path = os.path.join(self.TEST_IMG_DIR, 'best-fit-off-by-one-bug.png')
        crop = Crop(Box(x1=0, y1=0, x2=960, y2=915), img_path)
        size = Size('960', w=960, h=594)
        new_crop = size.fit_to_crop(crop)
        self.assertNotEqual(new_crop.box.h, 593, "Calculated best fit height is 1 pixel too short")
        self.assertEqual(new_crop.box.h, 594)

    def test_size_order_bug(self):
        img_path = os.path.join(self.TEST_IMG_DIR, 'size-order-bug.png')
        crop = Crop(Box(x1=160, y1=0, x2=800, y2=640), img_path)
        size = Size('650', w=650, min_h=250)
        new_crop = size.fit_to_crop(crop)
        self.assertGreaterEqual(new_crop.box.w, 650,
            "Calculated best fit (%d) didn't get required width (650)" % new_crop.box.w)



from .image import (
    get_image_extension, is_transparent, exif_orientation,
    correct_colorspace, is_animated_gif, has_animated_gif_support, process_image,
    smart_resize)
from .paths import get_upload_foldername
from .sizes import get_min_size
from .thumbs import set_as_auto_crop, unset_as_auto_crop
from . import jsonutils as json



from functools import update_wrapper

from logging import getLogger
import six

from django import http
from django.utils.decorators import classonlymethod


logger = getLogger('django.request')


class View(object):
    """
    Intentionally simple parent class for all views. Only implements
    dispatch-by-method and simple sanity checking.
    """

    http_method_names = ['get', 'post', 'put', 'delete', 'head', 'options', 'trace']

    def __init__(self, **kwargs):
        """
        Constructor. Called in the URLconf; can contain helpful extra
        keyword arguments, and other things.
        """
        # Go through keyword arguments, and either save their values to our
        # instance, or raise an error.
        for key, value in six.iteritems(kwargs):
            setattr(self, key, value)

    @classonlymethod
    def as_view(cls, **initkwargs):
        """
        Main entry point for a request-response process.
        """
        # sanitize keyword arguments
        for key in initkwargs:
            if key in cls.http_method_names:
                raise TypeError(u"You tried to pass in the %s method name as a "
                                u"keyword argument to %s(). Don't do that."
                                % (key, cls.__name__))
            if not hasattr(cls, key):
                raise TypeError(u"%s() received an invalid keyword %r" % (
                    cls.__name__, key))

        def view(request, *args, **kwargs):
            self = cls(**initkwargs)
            if hasattr(self, 'get') and not hasattr(self, 'head'):
                self.head = self.get
            return self.dispatch(request, *args, **kwargs)

        # take name and docstring from class
        update_wrapper(view, cls, updated=())

        # and possible attributes set by decorators
        # like csrf_exempt from dispatch
        update_wrapper(view, cls.dispatch, assigned=())
        return view

    def dispatch(self, request, *args, **kwargs):
        # Try to dispatch to the right method; if a method doesn't exist,
        # defer to the error handler. Also defer to the error handler if the
        # request method isn't on the approved list.
        if request.method.lower() in self.http_method_names:
            handler = getattr(self, request.method.lower(), self.http_method_not_allowed)
        else:
            handler = self.http_method_not_allowed
        self.request = request
        self.args = args
        self.kwargs = kwargs
        return handler(request, *args, **kwargs)

    def http_method_not_allowed(self, request, *args, **kwargs):
        allowed_methods = [m for m in self.http_method_names if hasattr(self, m)]
        logger.warning('Method Not Allowed (%s): %s', request.method, request.path,
            extra={
                'status_code': 405,
                'request': self.request
            }
        )
        return http.HttpResponseNotAllowed(allowed_methods)



import os
import shutil
import uuid

from django.conf import settings

PATH = os.path.split(__file__)[0]
ORIG_IMG_PATH = os.path.join(PATH, 'data')


class CropdusterTestCaseMediaMixin(object):

    def setUp(self):
        super(CropdusterTestCaseMediaMixin, self).setUp()

        random = uuid.uuid4().hex
        self.TEST_IMG_ROOT = os.path.join(settings.MEDIA_ROOT, random)
        self.TEST_IMG_DIR = os.path.join(self.TEST_IMG_ROOT, 'data')
        self.TEST_IMG_DIR_RELATIVE = os.path.join(random, 'data')

        # Create directory for test images
        shutil.copytree(ORIG_IMG_PATH, self.TEST_IMG_DIR)

    def tearDown(self):
        super(CropdusterTestCaseMediaMixin, self).tearDown()

        # Remove all generated images
        shutil.rmtree(self.TEST_IMG_ROOT, ignore_errors=True)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations

import cropduster.models
import cropduster.fields
import cropduster.settings


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Image',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('object_id', models.PositiveIntegerField(null=True, blank=True)),
                ('field_identifier', models.SlugField(default=b'', blank=True)),
                ('prev_object_id', models.PositiveIntegerField(null=True, blank=True)),
                ('width', models.PositiveIntegerField(null=True, blank=True)),
                ('height', models.PositiveIntegerField(null=True, blank=True)),
                ('image', cropduster.fields.CropDusterSimpleImageField(db_column=b'path', upload_to=cropduster.models.generate_filename, width_field=b'width', storage=cropduster.models.StrFileSystemStorage(), height_field=b'height', db_index=True)),
                ('date_created', models.DateTimeField(auto_now_add=True)),
                ('date_modified', models.DateTimeField(auto_now=True)),
                ('attribution', models.CharField(max_length=255, null=True, blank=True)),
                ('attribution_link', models.URLField(max_length=255, null=True, blank=True)),
                ('caption', models.TextField(null=True, blank=True)),
                ('content_type', models.ForeignKey(to='contenttypes.ContentType')),
            ],
            options={
                'db_table': '%s_image' % cropduster.settings.CROPDUSTER_DB_PREFIX,
            },
        ),
        migrations.CreateModel(
            name='StandaloneImage',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('md5', models.CharField(max_length=32)),
                ('image', cropduster.fields.CropDusterImageField(db_column=b'image', default=b'', editable=False, upload_to=b'/Users/fdintino/Sites/cms/assets/media', blank=True)),
            ],
            options={
                'db_table': '%s_standaloneimage' % cropduster.settings.CROPDUSTER_DB_PREFIX,
            },
        ),
        migrations.CreateModel(
            name='Thumb',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=255, db_index=True)),
                ('width', models.PositiveIntegerField(default=0, null=True, blank=True)),
                ('height', models.PositiveIntegerField(default=0, null=True, blank=True)),
                ('crop_x', models.PositiveIntegerField(null=True, blank=True)),
                ('crop_y', models.PositiveIntegerField(null=True, blank=True)),
                ('crop_w', models.PositiveIntegerField(null=True, blank=True)),
                ('crop_h', models.PositiveIntegerField(null=True, blank=True)),
                ('date_modified', models.DateTimeField(auto_now=True)),
                ('image', models.ForeignKey(related_name='+', blank=True, to='cropduster.Image', null=True)),
                ('reference_thumb', models.ForeignKey(related_name='auto_set', blank=True, to='cropduster.Thumb', null=True)),
            ],
            options={
                'db_table': '%s_thumb' % cropduster.settings.CROPDUSTER_DB_PREFIX,
            },
        ),
        migrations.AlterUniqueTogether(
            name='image',
            unique_together=set([('content_type', 'object_id', 'field_identifier')]),
        ),
    ]



import logging
import os
import tempfile
import subprocess

from cropduster.settings import CROPDUSTER_GIFSICLE_PATH


logger = logging.getLogger(__name__)


class GifsicleImage(object):

    def __init__(self, im):
        if not CROPDUSTER_GIFSICLE_PATH:
            raise Exception(
                "Cannot use GifsicleImage without the gifsicle binary in the PATH")

        self.pil_image = im
        self.size = im.size

        filename = getattr(im, 'filename', None)
        if not filename or not os.path.exists(filename):
            temp_file = tempfile.NamedTemporaryFile(suffix='.gif')
            filename = temp_file.name
            im.save(filename)

        self._filename = filename

        self.cmd_args = [CROPDUSTER_GIFSICLE_PATH, '-O3', '-I', '-I', '-w']
        self.crop_args = []
        self.resize_args = []

    @property
    def args(self):
        return self.cmd_args + self.crop_args + self.resize_args

    def crop(self, box):
        x1, y1, x2, y2 = box
        if x2 < x1:
            x2 = x1
        if y2 < y1:
            y2 = y1

        self.size = (x2 - x1, y2 - y1)
        self.crop_args = ['--crop', "%d,%d-%d,%d" % (x1, y1, x2, y2)]
        return self

    def resize(self, size, method):
        # Ignore method, PIL's algorithms don't match up
        self.resize_args = [
            "--resize-fit", "%dx%d" % size,
            "--resize-method", "lanczos3",
            "--resize-colors", "128",
        ]
        return self

    def save(self, output_filename, **kwargs):
        args = self.args + ['-o', output_filename, self._filename]
        proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = proc.communicate()
        logger.debug(err)
